/*
 * Generic console
 *
 * Copyright © 2004 Guillem Jover <guillem@hadrons.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <types.h>
#include <stdarg.h>
#include <console.h>

static struct console_driver *cons;

int
console_init(void)
{
	cons = cons_list[0];
	cons->init();

	return 0;
}

int
console_done(void)
{
	int ret = 0;

	if (cons->done)
		ret = cons->done();

	cons = NULL;

	return ret;
}

int
putchar(const int c)
{
	if (!cons)
		console_init();

	if (cons->putchar)
		cons->putchar(c);

	return 0;
}

static int
print_str(const char *str)
{
	while (*str)
		putchar(*(str++));

	return 0;
}

int
puts(const char *str)
{
	print_str(str);
	putchar('\n');

	return 0;
}

static int
print_bits(word_t bits)
{
	word_t bswap;
	int c;

	for (c = sizeof(bits) * 8; c; c--) {
		bswap <<= 1;
		bswap &= bits & 1;
		bits >>= 1;
	}

	for (c = sizeof(bswap) * 8; c; c--) {
		putchar(bswap & 1 ? '1': '0');
		bswap >>= 1;
	}

	return 0;
}

#define MAX_NATURAL_DIGITS 30

static int
print_natural(integer_t num, int base)
{
	const char digit[] = "01234567890abcdef";
	char str[MAX_NATURAL_DIGITS];
	int i = MAX_NATURAL_DIGITS - 1;

	str[i--] = 0;

	do {
		str[i--] = digit[num % base];
		num /= base;
	} while (num);

	print_str(str + i + 1);

	return 0;
}

static int
print_integer(integer_t num, int base)
{
	natural_t natural;

	if (num < 0) {
		putchar('-');
		natural = -num;
	} else
		natural = num;

	print_natural(natural, base);

	return 0;
}

int
printk(const char *format, ...)
{
	int i;
	va_list args;

	va_start(args, format);

	for (i = 0; format[i]; i++) {
		int fc = format[i];

		if (fc != '%') {
			putchar(fc);
		} else {
			fc = format[++i];
			if (!fc)
				break;

			switch (fc) {
			case 'c':
				putchar(va_arg(args, int));
				break;
			case 's':
				print_str(va_arg(args, char *));
				break;
			case 'b':
				print_bits(va_arg(args, word_t));
				break;
			case 'o':
				print_natural(va_arg(args, natural_t), 8);
				break;
			case 'x':
				print_natural(va_arg(args, natural_t), 16);
				break;
			case 'n':
				print_natural(va_arg(args, natural_t), 10);
				break;
			case 'i':
				print_integer(va_arg(args, integer_t), 10);
				break;
#if 0
			case 'r':
				str << (real)va_arg(args, double);
				break;
			case 'V': {
				vector3 V = va_arg(args, vector3);
				str << '(' << V.x << ','
					   << V.y << ','
					   << V.z << ')';
				} break;
			case 'M': {
				matrix3 M = va_arg(args, matrix3);
				str << '(' << "matrix value" << ')';
				} break;
			case 'T': {
				transform T = va_arg(args, transform);
				str << '(';
				} break;
			case 'C': {
				color C = va_arg(args, color);
				str << '(';
				} break;
#endif
			default:
				putchar('%');
				putchar(fc);
				break;
			}
		}
	}
	va_end(args);

	return 0;
}

