/*
 * Raster: VGA driver
 *
 * Copyright © 1998-2001, 2005 Guillem Jover <guillem@hadrons.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 */

#include <raster/raster.h>
#include <raster/vga.h>

#define VGA_GFX_MEM_ADDR	(void *)0xa0000

static u8_t *vga_mem;
static int max_x, max_y, colors;

/* BIOS mode 13h - 320x200x256.  */
static const vga_regs_t vga_mode_320x200x256_regs = {
	.max_x = 320,
	.max_y = 200,
	.colors = 256,
	.crt = { 0x5F, 0x4F, 0x50, 0x82, 0x54, 0x80, 0xBF, 0x1F, 0x00,
		 0x41, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x9C, 0x8E,
		 0x8F, 0x28, 0x40, 0x96, 0xB9, 0xA3 },
	.atc = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
		 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x41, 0x00,
		 0x0F, 0x00, 0x00 },
	.gfx = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x05, 0x0F, 0xFF },
	.seq = { 0x03, 0x01, 0x0F, 0x00, 0x0E },
	.msc = { 0x63 },
};

/* BIOS mode 12h - 640x480x16 */
static const vga_regs_t vga_mode_640x480x16_regs = {
	.max_x = 640,
	.max_y = 480,
	.colors = 16,
	.crt = { 0x5F, 0x4F, 0x50, 0x82, 0x54, 0x80, 0x0B, 0x3E, 0x00,
		 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xEA, 0x8C,
		 0xDF, 0x28, 0x00, 0xE7, 0x04, 0xE3 },
	.atc = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
		 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x01, 0x00,
		 0x0F, 0x00, 0x00 },
	.gfx = { 0x00, 0x0F, 0x00, 0x20, 0x00, 0x00, 0x05, 0x0F, 0xFF },
	.seq = { 0x03, 0x01, 0x0F, 0x00, 0x06 },
	.msc = { 0xE3 },
};

/* BIOS mode 03h - 80x25x16.  */
static const vga_regs_t vga_mode_80x25x16_regs = {
	.max_x = 80,
	.max_y = 25,
	.colors = 16,
	.crt = { 0x5B, 0x4F, 0x53, 0x37, 0x51, 0x5B, 0x6C, 0x1F, 0x00,
		 0x0D, 0x0a, 0x0c, 0x00, 0x00, 0x00, 0x00, 0x5E, 0x2B,
		 0x5D, 0x28, 0x0F, 0x5E, 0x0A, 0xA3 },
	.atc = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x14, 0x07, 0x38,
		 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F, 0x0A, 0x00,
		 0x0F, 0x00, 0x00 },
	.gfx = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x0E, 0x00, 0xFF },
	.seq = { 0x03, 0x01, 0x03, 0x00, 0x03 },
	.msc = { 0xA7 },
};


static bool
vga_init(void)
{
	int i;
#if 0
	const vga_regs_t *r = &vga_mode_640x480x16_regs;
#else
	const vga_regs_t *r = &vga_mode_320x200x256_regs;
#endif

	max_x = r->max_x;
	max_y = r->max_y;
	colors = r->colors;

	/* Disable video output.  */
	vga_atc_screen(0);

	vga_msc_out(r->msc[0]);

	/* Synchronous reset on.  */
	vga_seq_out(0x00, 0x01);

	vga_seq_out(0x01, r->seq[1] | 0x20);
	for (i = 2; i < VGA_SEQ_NREGS; i++)
		vga_seq_out(i, r->seq[i]);

	/* Synchronous reset off.  */
	vga_seq_out(0x00, 0x03);

	/* Deprotect CRT registers 0-7.  */
	vga_crt_out(0x11, vga_crt_in(0x11) & 0x7f);

	for (i = 0; i < VGA_CRT_NREGS; i++)
		vga_crt_out(i, r->crt[i]);

	for (i = 0; i < VGA_GFX_NREGS; i++)
		vga_gfx_out(i, r->gfx[i]);

	for (i = 0; i < VGA_ATC_NREGS; i++)
		vga_atc_out(i, r->atc[i]);

	/* Turn screen on.  */
	vga_seq_out(0x01, vga_seq_in(0x01) & 0xdf);

	/* Enable video output.  */
	vga_atc_screen(0x20);

	vga_mem = VGA_GFX_MEM_ADDR;

	return true;
}

static bool
vga_done(void)
{
	return true;
}

static void
vga_put_pixel(u32_t x, u32_t y, u8_t color)
{
	vga_mem[x + y * max_x] = color;
}

static u8_t
vga_get_pixel(u32_t x, u32_t y)
{
	return vga_mem[x + y * max_x];
}

static bool
vga_wait_vertical_retrace(void)
{
	while (vga_crt_status_in() & 8);
	while ((vga_crt_status_in() & 8) == 0);

	return true;
}

static void
vga_fill_screen(u8_t color)
{
	int i;

	for (i = 0; i < max_x * max_y; i++)
		vga_mem[i] = color;
}

static color_t
vga_get_pal_item(int index)
{
	color_t color;

	outb(VGA_DAC_REG_ADDR_R, index);
	color.r = inb(VGA_DAC_REG_DATA);
	color.g = inb(VGA_DAC_REG_DATA);
	color.b = inb(VGA_DAC_REG_DATA);

	return color;
}

static int
vga_put_pal_item(int index, color_t *color)
{
	outb(VGA_DAC_REG_ADDR_W, index);
	outb(VGA_DAC_REG_DATA, color->r);
	outb(VGA_DAC_REG_DATA, color->g);
	outb(VGA_DAC_REG_DATA, color->b);

	return true;
}

#if 0
static int
vga_get_pal(color_t *color)
{
	int i;

	for (i = 0; i <= 255; i++) {
		vga_get_pal_item(i, &color);
		color++;
	}

	return true;
}

static int
vga_put_pal(color_t *color)
{
	int i;

	for (i = 0; i <= 255; i++) {
		vga_put_pal_item(i, color);
		color++;
	}

	return true;
}

static int
vga_fade_out()
{
	u32_t x, y;

	for (x = 1; x <= 64; x++) {
		vga_wait_vertical_retrace();

		for (y = 0; y <= 255; y++) {
			color_t color = vga_get_pal_item(y);

			if (color.r > 0)
				color.r--;
			if (color.g > 0)
				color.g--;
			if (color.b > 0)
				color.b--;

			vga_put_pal_item(y, &color);
		}
	}

	return true;
}

static int
vga_fade_in(PaletaFinal:PaletaRGB)
{
	u32_t x, y;

	for (x = 1; x <= 64; x++) {
		vga_wait_vertical_retrace();

		for (y = 0; y <= 255; y++) {
			color_t color = vga_get_pal_item(y);

			if (color.r < paletafinal[y].r)
				color.r++;
			if (color.g < paletafinal[y].g)
				color.g++;
			if (color.b < paletafinal[y].b)
				color.b++;

			vga_put_pal_item(y, &color);
		}
	}
}
#endif

#if 0
Procedure RotaPal(var paleta:paletaRGB);
   var x:integer;
       tmp:colorRGB;
   begin
       EsperaVR;
       move(paleta[255],tmp,3);
       move(paleta[0],paleta[1],255*3);
       move(tmp,paleta[0],3);
       for x:=0 to 255 do PosaColPal(x,paleta[x].r,paleta[x].g,paleta[x].b);
   end;

Procedure Linia(x1,y1,x2,y2,Color:integer);
   Function Sgn(a:real):integer;
      begin
	 if a>0 then sgn:=+1;
	 if a<0 then sgn:=-1;
	 if a=0 then sgn:=0;
      end;

   var u,s,v,d1x,d1y,d2x,d2y,m,n:real;
       i:integer;
   begin
      u:=y2-x1;
      v:=x2-y1;
      d1x:=sgn(u);
      d1y:=sgn(v);
      d2x:=sgn(u);
      d2y:=0;
      m:=abs(u);
      n:=abs(v);
      if not (m>n) then begin
	 d2x:=0 ;
	 d2y:=sgn(v);
	 m:=abs(v);
	 n:=abs(u);
      end;
      s:=int(m/2);
      for i:=0 to round(m) do begin
	 PosaPixel(x1,y1,color);
	 s:=s+n;
	 if not (s<m) then begin
	    s:=s-m;
	    x1:=x1+round(d1x);
	    y1:=y1+round(d1y);
	 end else begin
	    x1:=x1+round(d2x);
	    y1:=y1+round(d2y);
	 end;
      end;
   end;

Procedure LiniaH( X, Y, Tamany: Integer; Color: Byte); Assembler;
   asm
      mov ax,$a000
      mov es,ax
      mov bx,[X]
      mov dx,[Y]
      xchg dh,dl
      mov di,dx
      shr di,2
      add di,dx
      add di,bx
      mov al,[Color]
      mov al,ah
      mov cx,[Tamany]
      rep stosb
   end;

Procedure LiniaV( X, Y, Tamany: Integer; Color: Byte); Assembler;
   asm
      mov ax,$a000
      mov es,ax
      mov bx,[X]
      mov dx,[Y]
      xchg dh,dl
      mov di,dx
      shr di,2
      add di,dx
      add di,bx
      mov al,[Color]
   @Bucle:
      mov cx,[Tamany]
      mov es:[di],al
      add di,200
      loop @Bucle
   end;

Procedure Cercle(Cx,Cy,Radi:integer;Color:Byte);
   var deg:real;
       x,y:integer;
   begin
      deg:=0;
      repeat
	 X:=round(radi*cos(deg));
	 Y:=round(radi*sin(deg));
	 posapixel(X+Cx,Y+Cy,color);
	 deg:=deg+0.005;
      until (deg>6.4);
   end;

Procedure CreaRectangle(x1,y1,x2,y2:integer;color:byte);
   var x,y:integer;
   begin
      for y:=y1 to y2 do for x:=x1 to x2 do PosaPixel(x,y,color);
   end;

Procedure CreaMarc(x1,y1,x2,y2:integer;color:byte);
   var i:integer;
   begin
      for i:=x1 to x2 do posapixel(i,y1,color);
      for i:=x1 to x2 do posapixel(i,y2,color);
      for i:=y1 to y2 do posapixel(x1,i,color);
      for i:=y1 to y2 do posapixel(x2,i,color);
   end;
#endif

static struct raster_driver vga_raster = {
	.name = "vga_raster",
	.init = vga_init,
	.done = vga_done,
	.wait_vertical_retrace = vga_wait_vertical_retrace,
	.put_pixel = vga_put_pixel,
	.get_pixel = vga_get_pixel,
	.put_pal_item = vga_put_pal_item,
	.get_pal_item = vga_get_pal_item,
	.fill_screen = vga_fill_screen,
};

struct raster_driver *raster_list[] = {
	&vga_raster,
	NULL,
};

