;
; 3D OS Loader code
;
;	Load the kernel image
;	Prepare machine
;	Start kernel
;
; Copyright © 1998-2001, 2004 Guillem Jover <guillem@hadrons.org>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software Foundation,
; Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
;

#define ASM
#include <boot/loader.h>
#include <boot/multiboot.h>
#include "loader_emu.asm"

    bits 16
    org 0

	jmp	short loader

	id	db '<3Dloader>v0.1 ', 0

loader:

#ifdef EMULATE_LOADER
	loader_emu_init
#else

	mov	ax, INIT_SEGMENT
	mov	es, ax
	xor	di, di
	mov	ax, LOADER_SEGMENT
	mov	ds, ax
	xor	si, si
	cli
	mov	ss, ax
	mov	sp, STACK_TOP - 2
	sti

	mov	si, id
	call	puts
	call	putnl

	cmp	[loader_sign], word LOADER_SIGN
	jz	loader_ok
	mov	si, loader_panic
	call	puts
	jmp	hang

loader_ok:

#endif

	mov	cx, LOADER_PARAMS_LEN	; initialize loader params
	mov	di, 0
	rep	stosb

	call	cpu_detect		; CPU detection to skip CPU < 386
	call	speed_detect		; CPU speed detection in MHz

	call	video_detect
	call	video_print_info

	; Load the kernel

load_elf_header:

	xor	ah, ah			; reset FDC
	xor	dl, dl
	int	0x13

	mov	ax, SYS_SEGMENT
	mov	es, ax
	xor	bx, bx
	mov	cx, 2 + LOADER_SECTORS
	xor	dx, dx
	mov	ah, 2
	mov	al, 1
	int	0x13
	jc	load_elf_header

	; Check for an ELF kernel

	mov	ax, SYS_SEGMENT
	mov	es, ax
	mov	eax, [elf_magic]
	cmp	eax, [es:0]
	je	elf_kernel
	mov	si, msg_elf_error
	call	puts
	jmp	hang

elf_kernel:

	mov	eax, [es:24]
	mov	[elf_entry], eax

	mov	si, msg_elf_entry_addr
	call	puts
	mov	dx, [elf_entry+4]
	call	putw
	mov	dx, [elf_entry]
	call	putw
	call	putnl

	mov	ax, [es:40]
	mov	[elf_size], ax
	shr	ax, 9
	mov	ax, 31		; XXX hardcode for now, need to compute
				; the actual size from the program header
	mov	[elf_sectors], ax

	mov	si, msg_elf_size
	call	puts
	mov	edx, [elf_size]
	call	putw
	call	putnl

	mov	si, msg_elf_loading
	call	puts

load_kernel:

	xor	ah, ah			; reset FDC
	xor	dl, dl
	int	0x13

	mov	al, '.'
	call	putc

	mov	ax, SYS_SEGMENT
	mov	es, ax
	mov	bx, 512
	mov	cx, 2 + LOADER_SECTORS + 1
	xor	dx, dx
	mov	ah, 2
	mov	al, [elf_sectors]
	int	0x13
	jc	load_kernel

	call	putnl

verify_entry:

	; Verify for a valid entry point

	; Normalize the address by subtracting the shifted segment as
	; elf_entry is a linear address.
	mov	eax, SYS_SEGMENT
	mov	es, ax
	mov	edi, [elf_entry]
	shl	eax, 4
	sub	edi, eax
	cmp	[es:edi], dword 0x90909090
	je	.valid_entry
	mov	si, msg_elf_entry_bad
	call	puts
	jmp	hang
.valid_entry:
	mov	si, msg_elf_entry_ok
	call	puts

	; Check for a multiboot kernel

	; Detect memory size

	; Detect APM BIOS

	mov	ax, 0x5300
	xor	bx, bx
	int	0x15
	jnc	apm_ok
	; error code on ah
	mov	si, msg_apm_error
	call	puts

apm_ok:
	; XXX: store apm data

	; Disable NMI and cpu interrupts

	cli
	cld
	mov	al, 0x80
	out	0x70, al
	in	al, 0x71

	; Enable A20 gate

	call	kb_empty_buffer
	mov	al, 0xD1			; Send command to i8042
	out	0x64, al
	call	kb_empty_buffer
	mov	al, 0xDF
	out	0x60, al
	call	kb_empty_buffer

	; Setup GDT and IDT

%define FLAT_CS_SELECTOR 1 * 8		; = 1 * sizeof(descriptor)
%define FLAT_DS_SELECTOR 2 * 8		; = 2 * sizeof(descriptor)

	lidt	[idt_desc]
	xor	eax, eax		; compute base GDT address (ds:gdt)
	mov	ax, ds
	shl	eax, 4
	add	eax, gdt
	mov	[gdt_desc + 2], eax
	lgdt	[gdt_desc]

	; Mask all interrupts in PIT

	; Flush prefech queue

	jmp short flush_queue
flush_queue:

	; Enable Protected Mode

	mov	eax, cr0
	mov	eax, 1
	mov	cr0, eax

	mov	esi, [elf_entry]

	; Jump to 32-bit protected mode

	jmp FLAT_CS_SELECTOR:dword ((LOADER_SEGMENT << 4) + prot_mode)

    bits 32

prot_mode:

	mov	ax, FLAT_DS_SELECTOR
	mov	ds, ax
	mov	es, ax
	mov	fs, ax
	mov	gs, ax
	mov	ss, ax
	mov	esp, 0x0000FFFE

	mov	eax, MULTIBOOT_BOOTLOADER_MAGIC
;	mov	ebx, boot_multiboot_data

	; Jump to kernel image (esi has entry point)

	jmp	esi

hang:
#ifdef EMULATE_LOADER
	loader_emu_end
#else
	jmp	hang
#endif

    bits 16

kernel_sectors	db KERNEL_SECTORS
elf_magic	db 177q, 'ELF'
elf_entry	dd 0
elf_sectors	dw 0
elf_size	dw 0
system_halt	db 'System halted...', 0
loader_panic	db 'Loader panic, not loaded correctly, halting...', 0
msg_apm_error	db 'No APM BIOS found.', 0
msg_elf_error	db 'No ELF image found.', 0
msg_elf_loading		db 'Loading ELF kernel ', 0
msg_elf_size		db 'ELF image size: ', 0
msg_elf_entry_addr	db 'ELF entry point: ', 0
msg_elf_entry_ok	db 'ELF entry point: valid', 0
msg_elf_entry_bad	db 'ELF entry point: invalid', 0


align 16

gdt:
  null_desc	dw 0, 0, 0, 0
  flat_cs_desc	dw 0xffff, 0, 0x9a00, 0x00cf
  flat_ds_desc	dw 0xffff, 0, 0x9200, 0x00cf
gdt_end:


align 4

idt_desc:	dw 0
		dd 0
gdt_desc:	dw gdt - gdt_end - 1
		dd 0			; Filled at runtime, with the linear
					; address


;
; Loader Library
;

kb_empty_buffer:
	push	ecx
	mov	ecx, 100000
.loop:
	dec	ecx
	jz	.loop_end
	call	delay
	in	al, 0x64
	test	al, 1
	jz	.no_output
	call	delay
	in	al, 0x60
	jmp	.loop
.no_output:
	test	al, 1
	jnz	.loop
.loop_end:
	pop	eax
	ret

delay:
	out	0x80, al
	ret

; function: puts(unsigned char *);
; input:
;    [DS:SI] -> ASCIZ to write

puts:
	push	ax
	push	bx
.print:	lodsb
	or	al, al
	je	.fi
	mov	ah, 0xE
	mov	bx, 0x0007
	int	0x10
	jmp	short .print
.fi:	pop	bx
	pop	ax
	ret

; function: putdec(unsigned);
; input:
;    AX -> word to write in decimal

putdec:
	push	bp
	mov	bp, sp
	sub	sp, 30
	push	es
	mov	bx, ds
	mov	es, bx
	mov	di, bp
	sub	di, 2
	std
	push	ax
	xor	al, al			; Mark \0 end of an ASCIZ
	stosb
	pop	ax
	mov	cx, 10
.next:	xor	dx, dx
	div	cx
	add	dx, 0x30
	xchg	ax, dx
	stosb
	xchg	ax, dx
	cmp	ax, 0
	jnz	.next
	cld
	pop	es
	mov	si, di
	inc	si
	call	puts
	mov	sp, bp
	pop	bp
	ret

; function: putw(unsigned);
; input:
;    DX -> word to write

putw:
	xchg	dl, dh
	call	putb
	xchg	dl, dh

; function: putb(unsigned char);
; input:
;    DL -> byte to write

putb:
	push	dx
	push	cx
	mov	cl, 4
	shr	dl, cl
	pop	cx
	call	putdigit
	pop	dx

; function: putdigit(unsigned char);
; input:
;    DL -> low nibble to write

putdigit:
	push	ax
	mov	ax, 0x0e0f
	and	al, dl
	add	al, 0x90
	daa
	add	al, 0x40
	daa
	int	0x10
	pop	ax
	ret

; function: putnl(void);

putnl:
	push	ax
	mov	al, 0xD
	call	putc
	mov	al, 0xA
	call	putc
	pop	ax
	ret

; function: putc(unsigned char);
; input:
;     AL -> Char to write

putc:
	push	bx
	push	bp
	push	dx
	mov	ah, 0xE
	int	0x10
	pop	dx
	pop	bp
	pop	bx
	ret

#include "cpu_id.asm"
#include "video.asm"

align 4

loader_sign       dw LOADER_SIGN

